/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/

#ifndef __SBSTRINGUTILS_H__
#define __SBSTRINGUTILS_H__

#include <nsStringAPI.h>
#include <nsTArray.h>
#include <prprf.h>

class nsIStringEnumerator;

/**
 * Class used to create strings from other data types.
 */
class sbAutoString : public nsAutoString
{
public:
  sbAutoString(int aValue)
  {
    char valueStr[64];

    PR_snprintf(valueStr, sizeof(valueStr), "%d", aValue);
    AssignLiteral(valueStr);
  }

  sbAutoString(PRUint32 aValue)
  {
    char valueStr[64];

    PR_snprintf(valueStr, sizeof(valueStr), "%lu", aValue);
    AssignLiteral(valueStr);
  }

  sbAutoString(PRUint64 aValue)
  {
    char valueStr[64];

    PR_snprintf(valueStr, sizeof(valueStr), "%llu", aValue);
    AssignLiteral(valueStr);
  }
};

/**
 * Class used to pass void strings to functions.
 *
 * E.g., SomeFunction(SBVoidString());
 */
class SBVoidString : public nsString
{
public:
  SBVoidString()
  {
    SetIsVoid(PR_TRUE);
  }
};

/// @see nsString::FindCharInSet
PRInt32 nsString_FindCharInSet(const nsAString& aString,
                               const char *aPattern,
                               PRInt32 aOffset = 0);

void AppendInt(nsAString& str, PRUint64 val);

PRInt64 nsString_ToInt64(const nsAString& str, nsresult* rv = nsnull);

PRUint64 nsString_ToUint64(const nsAString& str, nsresult* rv = nsnull);

nsresult SB_StringEnumeratorEquals(nsIStringEnumerator* aLeft,
                                   nsIStringEnumerator* aRight,
                                   PRBool* _retval);

/**
 * Searches a string for any occurences of any character of a set and replace
 * them with a replacement character.  Modifies the string in-place.
 * @see nsString_internal::ReplaceChar
 */
void nsString_ReplaceChar(/* inout */ nsAString& aString,
                          const nsAString& aOldChars,
                          const PRUnichar aNewChar);

/**
 * Return true if the given string is possibly UTF8
 * (i.e. it errs on the side of returning true)
 *
 * Note that it assumes all 7-bit encodings are utf8, and doesn't check for
 * invalid characters (e.g. 0xFFFE, surrogates).  This is a weaker check than
 * the nsReadableUtils version.
 */
PRBool IsLikelyUTF8(const nsACString& aString);

/**
 * Split the string specified by aString into sub-strings using the delimiter
 * specified by aDelimiter and place the sub-strings in the array specified by
 * aStringArray.
 *
 * \param aString               String to split.
 * \param aDelimiter            Sub-string delimiter.
 * \param aSubStringArray       Array of sub-strings.
 */
void nsString_Split(const nsAString&    aString,
                    const nsAString&    aDelimiter,
                    nsTArray<nsString>& aSubStringArray);

/**
 * Split the string specified by aString into sub-strings using the delimiter
 * specified by aDelimiter and place the sub-strings in the array specified by
 * aStringArray.
 *
 * \param aString               String to split.
 * \param aDelimiter            Sub-string delimiter.
 * \param aSubStringArray       Array of sub-strings.
 */
void nsCString_Split(const nsACString&    aString,
                     const nsACString&    aDelimiter,
                     nsTArray<nsCString>& aSubStringArray);

/*
 * Songbird string bundle URL.
 */
#define SB_STRING_BUNDLE_URL "chrome://songbird/locale/songbird.properties"

/**
 * Get and return in aString the localized string with the key specified by aKey
 * using the string bundle specified by aStringBundle.  If the string cannot be
 * found, return the default string specified by aDefault; if aDefault is not
 * specified, return aKey.
 *
 * If aStringBundle is not specified, use the main Songbird string bundle.
 *
 * \param aString               Returned localized string.
 * \param aKey                  Localized string key.
 * \param aDefault              Optional default string.
 * \param aStringBundle         Optional string bundle.
 */
nsresult SBGetLocalizedString(nsAString&             aString,
                              const nsAString&       aKey,
                              const nsAString&       aDefault,
                              class nsIStringBundle* aStringBundle = nsnull);

nsresult SBGetLocalizedString(nsAString&       aString,
                              const nsAString& aKey);

nsresult SBGetLocalizedString(nsAString&             aString,
                              const char*            aKey,
                              const char*            aDefault = nsnull,
                              class nsIStringBundle* aStringBundle = nsnull);

/**
 * Class used to create localized strings.
 *
 * E.g., SomeFunction(SBLocalizedString("string_key"));
 */
class SBLocalizedString : public nsString
{
public:
  SBLocalizedString(const nsAString&       aKey,
                    const nsAString&       aDefault,
                    class nsIStringBundle* aStringBundle = nsnull)
  {
    nsString stringValue;
    SBGetLocalizedString(stringValue, aKey, aDefault, aStringBundle);
    Assign(stringValue);
  }

  SBLocalizedString(const nsAString& aKey)
  {
    nsString stringValue;
    SBGetLocalizedString(stringValue, aKey);
    Assign(stringValue);
  }

  SBLocalizedString(const char*            aKey,
                    const char*            aDefault = nsnull,
                    class nsIStringBundle* aStringBundle = nsnull)
  {
    nsString stringValue;
    SBGetLocalizedString(stringValue, aKey, aDefault, aStringBundle);
    Assign(stringValue);
  }
};

template <class T>
inline
T const & sbAppendStringArrayDefaultExtractor(T const & aArrayItem)
{
  return aArrayItem;
}

template <class S, class Sep, class T, class E>
inline
S sbAppendStringArray(S & aTarget,
                      Sep const & aSeparator,
                      T const & aStringArray,
                      E aExtractor)
{
  // Save off the lengths
  PRUint32 const separatorLength = aSeparator.Length();
  PRUint32 const stringLength = aTarget.Length();
  PRUint32 const length = aStringArray.Length();

  // Calc the final length from the original string + length of all the
  // separators
  PRUint32 finalLength = stringLength + (separatorLength * (aStringArray.Length() - 1));
  PRUint32 index;
  for (PRUint32 index = 0; index < length; ++index) {
    finalLength += aStringArray[index].Length();
  }
  typename S::char_type const * const separator = aSeparator.BeginReading();
  // Set the writePosition to the end of the string
  typename S::char_type * writePosition;
  typename S::char_type * endPosition;
  aTarget.BeginWriting(&writePosition, &endPosition, finalLength);
  if (!writePosition)
    return aTarget;
  writePosition += stringLength;
  // Now append
  PRBool const isSeparatorEmpty = aSeparator.IsEmpty();
  for (PRUint32 index = 0; index < length; ++index) {
    if (index != 0 && !isSeparatorEmpty) {
      memcpy(writePosition, separator, separatorLength * sizeof(typename S::char_type));
      writePosition += separatorLength;
    }
    S const & value = aExtractor(aStringArray[index]);
    memcpy(writePosition,
           value.BeginReading(),
           value.Length() * sizeof(typename S::char_type));
    writePosition += value.Length();
  }
  return aTarget;
}

template <class S, class Sep, class T>
inline
S sbAppendStringArray(S & aTarget,
                      Sep const & aSeparator,
                      T const & aStringArray)
{
  return sbAppendStringArray(aTarget, aSeparator, aStringArray, sbAppendStringArrayDefaultExtractor<S>);
}

#endif /* __SBSTRINGUTILS_H__ */

